
#include <zlib.h>
#include "dxa_player.h"
#include "systemstub.h"


bool DXA_Player::open(const char *filename) {
	bool opened = false;
	printf("opening '%s'\n", filename);
	if (_fd.open(filename, "rb")) {
		uint32 tag = _fd.readUint32BE();
		if (tag == _tagDEXA) {
			_fd.readByte();
			_framesCount = _fd.readUint16BE();
			_frameTicks = _fd.readUint32BE();
			if (_frameTicks > 100) {
				_frameTicks = 100;
			}
			_width = _fd.readUint16BE();
			_height = _fd.readUint16BE();
			printf("frames_count %d width %d height %d ticks %d\n", _framesCount, _width, _height, _frameTicks);
			_frameSize = _width * _height;
			_frameBuffer1 = (uint8 *)malloc(_frameSize);
			_frameBuffer2 = (uint8 *)malloc(_frameSize);
			if (!_frameBuffer1 || !_frameBuffer2) {
				printf("error allocating frame tables, size %d\n", _frameSize);
				close();
			} else {
				tag = _fd.readUint32BE();
				if (tag == _tagWAVE) {
					uint32 size = _fd.readUint32BE();
					printf("wave_size = %d\n", size);
					// XXX preload wave data
					_fd.seek(size);
				}
				_currentFrame = 0;
				sysInit(_width, _height, filename);
				opened = true;
			}
		}
	}
	return opened;
}

void DXA_Player::close() {
	sysDestroy();
	_fd.close();
	free(_frameBuffer1);
	free(_frameBuffer2);
}

void DXA_Player::play() {
	while (_currentFrame < _framesCount) {
		handleNextFrame();
		++_currentFrame;
	}
}

void DXA_Player::handleNextFrame() {
	uint32 tag = _fd.readUint32BE();
	if (tag == _tagCMAP) {
		uint8 pal[256 * 3];
		_fd.read(pal, ARRAYSIZE(pal));
		sysSetPalette(pal, 256);
	}
	tag = _fd.readUint32BE();
	if (tag == _tagFRAM) {
		uint8 type = _fd.readByte();
		uint32 size = _fd.readUint32BE();
		printf("frame %d type %d size %d\n", _currentFrame, type, size);
		_fd.read(_frameBuffer2, size);
		switch (type) {
		case 2:
		case 3:
			decodeZlib(_frameBuffer2, size, _frameSize);
			break;
		case 4:
		case 5:
			decode0(_frameBuffer2, size);
			break;
		case 6:
		case 7:
			decode2(_frameBuffer2, size, _frameSize);
			break;
		case 12:
			decode12(_frameBuffer2, size, _frameSize);
		}
		if (type == 2 || type == 4 || type == 6 || type == 12) {
			memcpy(_frameBuffer1, _frameBuffer2, _frameSize);
		} else {
			for (int j = 0; j < _height; ++j) {
				for (int i = 0; i < _width; ++i) {
					const int offs = j * _width + i;
					_frameBuffer1[offs] ^= _frameBuffer2[offs];
				}
			}
		}
	}
	sysCopyRect(_frameBuffer1, _width, 0, 0, _width, _height);
	if (sysDelay(30)) exit(0);
}

void DXA_Player::decodeZlib(uint8 *data, int size, int totalSize) {
	uint8 *temp = (uint8 *)malloc(size);
	if (temp) {
		memcpy(temp, data, size);
    	z_stream d_stream;
    	d_stream.zalloc = (alloc_func)0;
    	d_stream.zfree = (free_func)0;
    	d_stream.opaque = (voidpf)0;
    	d_stream.next_in = temp;
    	d_stream.avail_in = size;
    	d_stream.total_in = size;
    	d_stream.next_out = data;
    	d_stream.avail_out = totalSize;
    	inflateInit(&d_stream);
        inflate(&d_stream, Z_FINISH);
    	inflateEnd(&d_stream);
		free(temp);
	}
}

void DXA_Player::decode0(uint8 *data, int size) {
	// XXX
}

void DXA_Player::decode2(uint8 *data, int size, int totalSize) {
	// XXX
}

void DXA_Player::decode12(uint8 *data, int size, int totalSize) {
	// XXX

	#define BLOCKW 4
	#define BLOCKH 4

	/* decompress the input data */
	decodeZlib(data, size, totalSize);

	uint8 *dat = data;
	uint8 *frame2 = (uint8*)malloc(totalSize);

	memcpy(frame2, _frameBuffer1, totalSize);

	
        for (int by = 0; by < _height; by += BLOCKH) {
                for (int bx = 0; bx < _width; bx += BLOCKW) {
                        uint8 type = *dat++;
                        
                        uint8 *b2 = (uint8*)frame2 + bx + by * _width;
                        
                        switch (type) {
                                case 0:
                                        break;
								case 10:
								case 11:
								case 12:
								case 13:
								case 14:
								case 15:
                                case 1:
                                        {
												unsigned short diffMap;
												if (type >= 10 && type <= 15) {
														static const struct { uint8 sh1, sh2; } shiftTbl[6] = {
															{0, 0},	{8, 0},	{8, 8},	{8, 4},	{4, 0},	{4, 4}
														};
														diffMap = ((*dat & 0xF0) << shiftTbl[type-10].sh1) | 
																  ((*dat & 0x0F) << shiftTbl[type-10].sh2);
														dat++;
												} else {
														diffMap = *(unsigned short*)dat;
														dat += 2;
												}
                                                
                                                for (int yc = 0; yc < BLOCKH; yc++) {
                                                        for (int xc = 0; xc < BLOCKW; xc++) {
                                                                if (diffMap & 0x8000) {
                                                                        b2[xc] = *dat++;
                                                                }
                                                                diffMap <<= 1;
                                                        }
                                                        b2 += _width;
                                                }
                                                
                                                break;
                                        }
                                case 2:
                                        {
                                                uint8 color = *dat++;
                                                
                                                for (int yc = 0; yc < BLOCKH; yc++) {
                                                        for (int xc = 0; xc < BLOCKW; xc++) {
                                                                b2[xc] = color;
                                                        }
                                                        b2 += _width;
                                                }
                                                
                                                break;
                                        }
                                case 3:
                                        {
                                                for (int yc = 0; yc < BLOCKH; yc++) {
                                                        for (int xc = 0; xc < BLOCKW; xc++) {
                                                                b2[xc] = *dat++;
                                                        }
                                                        b2 += _width;
                                                }
                                                
                                                break;
                                        }
                                        break;
                                case 4:
                                        {
                                                uint8 mbyte = *dat++;
                                                int mx = (mbyte >> 4) & 0x07;
                                                if (mbyte & 0x80) mx = -mx;
                                                int my = mbyte & 0x07;
                                                if (mbyte & 0x08) my = -my;
												uint8 *b1 = (uint8*)_frameBuffer1 + (bx+mx) + (by+my) * _width;
                                                for (int yc = 0; yc < BLOCKH; yc++) {
                                                        memcpy(b2, b1, BLOCKW);
                                                        b1 += _width;
                                                        b2 += _width;
                                                }

										}
                                        break;
                                case 5:
                                        break;
                        }
				}
		}
	

	memcpy(data, frame2, totalSize);
	free(frame2);

}

const uint32 DXA_Player::_tagDEXA = 0x44455841;

const uint32 DXA_Player::_tagFRAM = 0x4652414D;

const uint32 DXA_Player::_tagWAVE = 0x57415645;

const uint32 DXA_Player::_tagCMAP = 0x434D4150;
